#
# Copyright 2021-2025 WhiteMech
#
# ------------------------------
#
# This file is part of pddl.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
#

"""Package metadata declaration."""

__title__ = "pddl"
__description__ = "PDDL parser"
__url__ = "https://github.com/AI-Planning/pddl.git"
__version__ = "0.4.7"
__author__ = "Marco Favorito, Francesco Fuggitti, Christian Muise"
__author_email__ = (
    "marco.favorito@gmail.com, francesco.fuggitti@gmail.com, christian.muise@queensu.ca"
)
__license__ = "MIT License"
__copyright__ = "2021-2025 WhiteMech"
